==> ir_sim-2.2.5.tar.gz: ir_sim-2.2.5/irsim/env/env_plot.py <==
import matplotlib.pyplot as plt
import logging
from irsim.global_param.path_param import path_manager as pm
from irsim.global_param import world_param, env_param
import os
import imageio
import shutil
import glob
from math import sin, cos
import numpy as np
import matplotlib
matplotlib.use("TkAgg")


class EnvPlot:
    """
    EnvPlot class for visualizing the environment.

    Args:
        grid_map (optional): The grid map of the environment (PNG file).
        objects: List of objects in the environment.
        x_range (list): The range of x-axis values. Default is [0, 10].
        y_range (list): The range of y-axis values. Default is [0, 10].
        saved_figure (dict): Keyword arguments for saving the figure.
            See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.savefig.html for details.
        saved_ani (dict): Keyword arguments for saving the animation.
            See https://imageio.readthedocs.io/en/v2.8.0/format_gif-pil.html#gif-pil for details.
        dpi: Dots per inch for the figure. Default is 100.
        figure_pixels: Width and height of the figure in pixels. Default is [1920, 1080].
        kwargs: Additional options such as color_map, no_axis, and tight.
    """

    def __init__(
        self,
        grid_map=None,
        objects=[],
        x_range=[0, 10],
        y_range=[0, 10],
        saved_figure=dict(),
        saved_ani=dict(),
        dpi: int = 100,
        figure_pixels: list =[1920, 1080],
        **kwargs,
    ) -> None:
        """
        Initialize the EnvPlot instance.
        """

        self.fig, self.ax = plt.subplots(figsize=(figure_pixels[0] / dpi, figure_pixels[1] / dpi), dpi=dpi)
        self.x_range = x_range
        self.y_range = y_range

        self.init_plot(grid_map, objects, **kwargs)
        self.color_map = {
            "robot": "g",
            "obstacle": "k",
            "landmark": "b",
            "target": "pink",
        }

        self.color_map.update(kwargs.get("color_map", dict()))

        self.saved_figure_kwargs = saved_figure
        self.saved_ani_kwargs = saved_ani
        self.dpi = dpi
        self.figure_pixels = figure_pixels

        self.dyna_line_list = []
        self.dyna_point_list = []

    def init_plot(self, grid_map, objects, no_axis=False, tight=True):
        """
        Initialize the plot with the given grid map and objects.

        Args:
            grid_map (optional): The grid map of the environment.
            objects (list): List of objects to plot.
            no_axis (bool, optional): Whether to show the axis. Default is False.
            tight (bool, optional): Whether to show the axis tightly. Default is True.
        """
        self.ax.set_aspect("equal")
        self.ax.set_xlim(self.x_range)
        self.ax.set_ylim(self.y_range)

        self.ax.set_xlabel("x [m]")
        self.ax.set_ylabel("y [m]")

        self.draw_components("static", objects)
        self.draw_grid_map(grid_map)

        if no_axis:
            plt.axis("off")
        if tight:
            self.fig.tight_layout()

    def draw_components(self, mode="all", objects=[], **kwargs):
        """
        Draw the components in the environment.

        Args:
            mode (str): 'static', 'dynamic', or 'all' to specify which objects to draw.
            objects (list): List of objects to draw.
            kwargs: Additional plotting options.
        """
        if mode == "static":
            [obj.plot(self.ax, **kwargs) for obj in objects if obj.static]
        elif mode == "dynamic":
            [obj.plot(self.ax, **kwargs) for obj in objects if not obj.static]
        elif mode == "all":
            [obj.plot(self.ax, **kwargs) for obj in objects]
        else:
            logging.error("Error: Invalid draw mode")

    def clear_components(self, mode="all", objects=[]):
        """
        Clear the components in the environment.

        Args:
            mode (str): 'static', 'dynamic', or 'all' to specify which objects to clear.
            objects (list): List of objects to clear.
        """
        if mode == "dynamic":
            [obj.plot_clear() for obj in objects if not obj.static]
            [line.pop(0).remove() for line in self.dyna_line_list]
            [points.remove() for points in self.dyna_point_list]

            self.dyna_line_list = []
            self.dyna_point_list = []

        elif mode == "static":
            pass

        elif mode == "all":
            if objects:
                [obj.plot_clear() for obj in objects]
            else:
                plt.cla()

    def draw_grid_map(self, grid_map=None, **kwargs):
        """
        Draw the grid map on the plot.

        Args:
            grid_map (optional): The grid map to draw.
        """
        if grid_map is not None:
            self.ax.imshow(
                grid_map.T,
                cmap="Greys",
                origin="lower",
                extent=self.x_range + self.y_range,
            )

    def draw_trajectory(
        self,
        traj,
        traj_type="g-",
        label="trajectory",
        show_direction=False,
        refresh=False,
        **kwargs,
    ):
        """
        Draw a trajectory on the plot.

        Args:
            traj (list or np.ndarray): List of points or array of points [x, y, theta].
            traj_type (str): Type of trajectory line (e.g., 'g-').
                See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.plot.html for details.
            label (str): Label for the trajectory.
            show_direction (bool): Whether to show the direction of the trajectory.
            refresh (bool): Whether to refresh the plot.
            kwargs: Additional plotting options.
        """
        if isinstance(traj, list):
            path_x_list = [p[0, 0] for p in traj]
            path_y_list = [p[1, 0] for p in traj]
        elif isinstance(traj, np.ndarray):
            path_x_list = [p[0] for p in traj.T]
            path_y_list = [p[1] for p in traj.T]

        line = self.ax.plot(path_x_list, path_y_list, traj_type, label=label, **kwargs)

        if show_direction:
            if isinstance(traj, list):
                u_list = [cos(p[2, 0]) for p in traj]
                y_list = [sin(p[2, 0]) for p in traj]
            elif isinstance(traj, np.ndarray):
                u_list = [cos(p[2]) for p in traj.T]
                y_list = [sin(p[2]) for p in traj.T]

            self.ax.quiver(path_x_list, path_y_list, u_list, y_list)

        if refresh:
            self.dyna_line_list.append(line)

    def draw_points(self, point_list, s=10, c="m", refresh=True, **kwargs):
        """
        Draw points on the plot.

        Args:
            point_list (list): List of points, each point as [x, y].
            s (int): Size of the points.
            c (str): Color of the points.
            refresh (bool): Whether to refresh the plot.
            kwargs: Additional plotting options.
        """
        if point_list is not None:
            x_coordinates = [point[0] for point in point_list]
            y_coordinates = [point[1] for point in point_list]

            points = self.ax.scatter(x_coordinates, y_coordinates, s, c, **kwargs)

            if refresh:
                self.dyna_point_list.append(points)

    def draw_box(self, vertices, refresh=False, color="b-"):
        """
        Draw a box by the vertices.

        Args:
            vertices (np.ndarray): 2xN array of vertices.
            refresh (bool): Whether to refresh the plot.
            color (str): Color and line type of the box.
        """
        temp_vertex = np.c_[vertices, vertices[0:2, 0]]
        box_line = self.ax.plot(temp_vertex[0, :], temp_vertex[1, :], color)

        if refresh:
            self.dyna_line_list.append(box_line)

    def save_gif_figure(self, format="png", **kwargs):
        """
        Save the figure for generating animation.

        Args:
            format (str): Format of the figure. Default is 'png'.
            kwargs: Additional arguments for saving the figure.
                See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.savefig.html for details.
        """
        fp = pm.ani_buffer_path

        if not os.path.exists(fp):
            os.makedirs(fp)

        order = str(world_param.count).zfill(3)

        self.saved_figure_kwargs.update({"dpi": self.dpi, "bbox_inches": "tight"})
        self.saved_figure_kwargs.update(kwargs)

        self.fig.savefig(
            fp + "/" + order + "." + format, format=format, **self.saved_figure_kwargs
        )
    
    def save_figure(self, format="png", save_name='', **kwargs):
        """
        Save the current figure.

        Args:
            format (str): Format of the figure. Default is 'png'.
            kwargs: Additional arguments for saving the figure.
                See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.savefig.html for details.
        """
        fp = pm.fig_path

        if not os.path.exists(fp):
            os.makedirs(fp)

        order = str(world_param.count).zfill(3)

        self.saved_figure_kwargs.update({"dpi": self.dpi, "bbox_inches": "tight"})
        self.saved_figure_kwargs.update(kwargs)

        self.fig.savefig(
            fp + "/" + save_name + '_' + order + "." + format, format=format, **self.saved_figure_kwargs
        )
    
    def save_animate(
        self,
        ani_name="animation",
        suffix=".gif",
        keep_len=30,
        rm_fig_path=True,
        **kwargs,
    ):
        """
        Save the animation.

        Args:
            ani_name (str): Name of the animation. Default is 'animation'.
            suffix (str): Suffix of the animation file. Default is '.gif'.
            keep_len (int): Length of the last frame. Default is 30.
            rm_fig_path (bool): Whether to remove the figure path after saving. Default is True.
            kwargs: Additional arguments for saving the animation.
                See https://imageio.readthedocs.io/en/v2.8.0/format_gif-pil.html for details.
        """
        self.saved_ani_kwargs.update({"subrectangles": True})
        self.saved_ani_kwargs.update(kwargs)

        env_param.logger.info("Start to create animation")

        ap = pm.ani_path
        fp = pm.ani_buffer_path

        if not os.path.exists(ap):
            os.makedirs(ap)

        images = list(glob.glob(fp + "/*.png"))

        images.sort()
        image_list = []
        for i, file_name in enumerate(images):
            if i == 0:
                continue

            image_list.append(imageio.imread(str(file_name)))
            if i == len(images) - 1:
                for j in range(keep_len):
                    image_list.append(imageio.imread(str(file_name)))

        imageio.mimsave(
            ap + "/" + ani_name + suffix, image_list, **self.saved_ani_kwargs
        )

        env_param.logger.info("Animation created successfully, saved in " + ap)

        if rm_fig_path:
            shutil.rmtree(fp)

    def show(self):
        """
        Display the plot.
        """
        plt.show()

==> ir_sim-2.2.6.tar.gz: ir_sim-2.2.6/irsim/env/env_plot.py <==
import matplotlib.pyplot as plt
import logging
from irsim.global_param.path_param import path_manager as pm
from irsim.global_param import world_param, env_param
import os
import imageio
import shutil
import glob
from math import sin, cos
import numpy as np
import matplotlib

matplotlib.use("TkAgg")


class EnvPlot:
    """
    EnvPlot class for visualizing the environment.

    Args:
        grid_map (optional): The grid map of the environment (PNG file).
        objects: List of objects in the environment.
        x_range (list): The range of x-axis values. Default is [0, 10].
        y_range (list): The range of y-axis values. Default is [0, 10].
        saved_figure (dict): Keyword arguments for saving the figure.
            See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.savefig.html for details.
        saved_ani (dict): Keyword arguments for saving the animation.
            See https://imageio.readthedocs.io/en/v2.8.0/format_gif-pil.html#gif-pil for details.
        dpi: Dots per inch for the figure. Default is 100.
        figure_pixels: Width and height of the figure in pixels. Default is [1920, 1080].
        kwargs: Additional options such as color_map, no_axis, and tight.
    """

    def __init__(
        self,
        grid_map=None,
        objects=[],
        x_range=[0, 10],
        y_range=[0, 10],
        saved_figure=dict(),
        saved_ani=dict(),
        dpi: int = 100,
        figure_pixels: list =[1920, 1080],
        **kwargs,
    ) -> None:
        """
        Initialize the EnvPlot instance.
        """

        self.fig, self.ax = plt.subplots(figsize=(figure_pixels[0] / dpi, figure_pixels[1] / dpi), dpi=dpi)
        self.x_range = x_range
        self.y_range = y_range

        self.init_plot(grid_map, objects, **kwargs)
        self.color_map = {
            "robot": "g",
            "obstacle": "k",
            "landmark": "b",
            "target": "pink",
        }

        self.color_map.update(kwargs.get("color_map", dict()))

        self.saved_figure_kwargs = saved_figure
        self.saved_ani_kwargs = saved_ani
        self.dpi = dpi
        self.figure_pixels = figure_pixels

        self.dyna_line_list = []
        self.dyna_point_list = []

    def init_plot(self, grid_map, objects, no_axis=False, tight=True):
        """
        Initialize the plot with the given grid map and objects.

        Args:
            grid_map (optional): The grid map of the environment.
            objects (list): List of objects to plot.
            no_axis (bool, optional): Whether to show the axis. Default is False.
            tight (bool, optional): Whether to show the axis tightly. Default is True.
        """
        self.ax.set_aspect("equal")
        self.ax.set_xlim(self.x_range)
        self.ax.set_ylim(self.y_range)

        self.ax.set_xlabel("x [m]")
        self.ax.set_ylabel("y [m]")

        self.draw_components("static", objects)
        self.draw_grid_map(grid_map)

        if no_axis:
            plt.axis("off")
        if tight:
            self.fig.tight_layout()

    def draw_components(self, mode="all", objects=[], **kwargs):
        """
        Draw the components in the environment.

        Args:
            mode (str): 'static', 'dynamic', or 'all' to specify which objects to draw.
            objects (list): List of objects to draw.
            kwargs: Additional plotting options.
        """
        if mode == "static":
            [obj.plot(self.ax, **kwargs) for obj in objects if obj.static]
        elif mode == "dynamic":
            [obj.plot(self.ax, **kwargs) for obj in objects if not obj.static]
        elif mode == "all":
            [obj.plot(self.ax, **kwargs) for obj in objects]
        else:
            logging.error("Error: Invalid draw mode")

    def clear_components(self, mode="all", objects=[]):
        """
        Clear the components in the environment.

        Args:
            mode (str): 'static', 'dynamic', or 'all' to specify which objects to clear.
            objects (list): List of objects to clear.
        """
        if mode == "dynamic":
            [obj.plot_clear() for obj in objects if not obj.static]
            [line.pop(0).remove() for line in self.dyna_line_list]
            [points.remove() for points in self.dyna_point_list]

            self.dyna_line_list = []
            self.dyna_point_list = []

        elif mode == "static":
            pass

        elif mode == "all":
            if objects:
                [obj.plot_clear() for obj in objects]
            else:
                plt.cla()

    def draw_grid_map(self, grid_map=None, **kwargs):
        """
        Draw the grid map on the plot.

        Args:
            grid_map (optional): The grid map to draw.
        """
        if grid_map is not None:
            self.ax.imshow(
                grid_map.T,
                cmap="Greys",
                origin="lower",
                extent=self.x_range + self.y_range,
            )

    def draw_trajectory(
        self,
        traj,
        traj_type="g-",
        label="trajectory",
        show_direction=False,
        refresh=False,
        **kwargs,
    ):
        """
        Draw a trajectory on the plot.

        Args:
            traj (list or np.ndarray): List of points or array of points [x, y, theta].
            traj_type (str): Type of trajectory line (e.g., 'g-').
                See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.plot.html for details.
            label (str): Label for the trajectory.
            show_direction (bool): Whether to show the direction of the trajectory.
            refresh (bool): Whether to refresh the plot.
            kwargs: Additional plotting options.
        """
        if isinstance(traj, list):
            path_x_list = [p[0, 0] for p in traj]
            path_y_list = [p[1, 0] for p in traj]
        elif isinstance(traj, np.ndarray):
            path_x_list = [p[0] for p in traj.T]
            path_y_list = [p[1] for p in traj.T]

        line = self.ax.plot(path_x_list, path_y_list, traj_type, label=label, **kwargs)

        if show_direction:
            if isinstance(traj, list):
                u_list = [cos(p[2, 0]) for p in traj]
                y_list = [sin(p[2, 0]) for p in traj]
            elif isinstance(traj, np.ndarray):
                u_list = [cos(p[2]) for p in traj.T]
                y_list = [sin(p[2]) for p in traj.T]

            self.ax.quiver(path_x_list, path_y_list, u_list, y_list)

        if refresh:
            self.dyna_line_list.append(line)

    def draw_points(self, point_list, s=10, c="m", refresh=True, **kwargs):
        """
        Draw points on the plot.

        Args:
            point_list (list): List of points, each point as [x, y].
            s (int): Size of the points.
            c (str): Color of the points.
            refresh (bool): Whether to refresh the plot.
            kwargs: Additional plotting options.
        """
        if point_list is not None:
            x_coordinates = [point[0] for point in point_list]
            y_coordinates = [point[1] for point in point_list]

            points = self.ax.scatter(x_coordinates, y_coordinates, s, c, **kwargs)

            if refresh:
                self.dyna_point_list.append(points)

    def draw_box(self, vertices, refresh=False, color="b-"):
        """
        Draw a box by the vertices.

        Args:
            vertices (np.ndarray): 2xN array of vertices.
            refresh (bool): Whether to refresh the plot.
            color (str): Color and line type of the box.
        """
        temp_vertex = np.c_[vertices, vertices[0:2, 0]]
        box_line = self.ax.plot(temp_vertex[0, :], temp_vertex[1, :], color)

        if refresh:
            self.dyna_line_list.append(box_line)

    def save_figure(self, file_name='', file_format="png", include_index=False, save_gif=False, **kwargs):
        """
        Save the current figure.

        Args:
            file_name (str): Name of the figure. Default is ''.
            file_format (str): Format of the figure. Default is 'png'.
            kwargs: Additional arguments for saving the figure.
                See https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.savefig.html for details.
        """

        if save_gif:
            fp = pm.ani_buffer_path
        else:
            fp = pm.fig_path

        if not os.path.exists(fp):
            os.makedirs(fp)

        self.saved_figure_kwargs.update({"dpi": self.dpi, "bbox_inches": "tight"})
        self.saved_figure_kwargs.update(kwargs)

        if include_index or save_gif:
            order = str(world_param.count).zfill(3)
            full_name = fp + "/" + file_name + '_' + order + "." + file_format
        else:
            full_name = fp + "/" + file_name + "." + file_format

        self.fig.savefig(
            full_name, format=file_format, **self.saved_figure_kwargs
        )
    
    def save_animate(
        self,
        ani_name="animation",
        suffix=".gif",
        keep_len=30,
        rm_fig_path=True,
        **kwargs,
    ):
        """
        Save the animation.

        Args:
            ani_name (str): Name of the animation. Default is 'animation'.
            suffix (str): Suffix of the animation file. Default is '.gif'.
            keep_len (int): Length of the last frame. Default is 30.
            rm_fig_path (bool): Whether to remove the figure path after saving. Default is True.
            kwargs: Additional arguments for saving the animation.
                See https://imageio.readthedocs.io/en/v2.8.0/format_gif-pil.html for details.
        """
        self.saved_ani_kwargs.update({"subrectangles": True})
        self.saved_ani_kwargs.update(kwargs)

        env_param.logger.info("Start to create animation")

        ap = pm.ani_path
        fp = pm.ani_buffer_path

        if not os.path.exists(ap):
            os.makedirs(ap)

        images = list(glob.glob(fp + "/*.png"))

        images.sort()
        image_list = []
        for i, file_name in enumerate(images):
            if i == 0:
                continue

            image_list.append(imageio.imread(str(file_name)))
            if i == len(images) - 1:
                for j in range(keep_len):
                    image_list.append(imageio.imread(str(file_name)))

        imageio.mimsave(
            ap + "/" + ani_name + suffix, image_list, **self.saved_ani_kwargs
        )

        env_param.logger.info("Animation created successfully, saved in " + ap)

        if rm_fig_path:
            shutil.rmtree(fp)

    def show(self):
        """
        Display the plot.
        """
        plt.show()


def linewidth_from_data_units(linewidth, axis, reference='y'):
    """
    Convert a linewidth in data units to linewidth in points.

    Parameters
    ----------
    linewidth: float
        Linewidth in data units of the respective reference-axis
    axis: matplotlib axis
        The axis which is used to extract the relevant transformation
        data (data limits and size must not change afterwards)
    reference: string
        The axis that is taken as a reference for the data width.
        Possible values: 'x' and 'y'. Defaults to 'y'.

    Returns
    -------
    linewidth: float
        Linewidth in points
    """
    fig = axis.get_figure()
    if reference == 'x':
        length = fig.bbox_inches.width * axis.get_position().width
        value_range = np.diff(axis.get_xlim())
    elif reference == 'y':
        length = fig.bbox_inches.height * axis.get_position().height
        value_range = np.diff(axis.get_ylim())
    # Convert length to points
    length *= 72
    # Scale linewidth to value range
    return linewidth * (length / value_range)
